import numpy as np
import os
import os.path as p
import shutil
import cPickle as pickle
import tempfile as tmp
import fcntl
import json

def cprint(x):
  pass

import os
from contextlib import contextmanager
from tempfile   import NamedTemporaryFile

if not hasattr(os, 'replace'):
    os.replace = os.rename #NOTE: it won't work for existing files on Windows

@contextmanager
def FaultTolerantFile(name):
    dirpath, filename = os.path.split(name)
    # use the same dir for os.rename() to work
    with NamedTemporaryFile(dir=dirpath, prefix=filename, suffix='.tmp') as f:
        yield f
        f.flush()   # libc -> OS
        os.fsync(f) # OS -> disc (note: on OSX it is not enough)
        f.delete = False # don't delete tmp file if `replace()` fails
        f.close()
        os.replace(f.name, name)
        
        

class Folder(object):
  def __init__(self,folders=None,create=False,rm=False):
    self._create = create
    
    if folders is None:
      # create empty tmp folder
      self._path = tmp.mkdtemp()
        
    else:
      # folder is single path string
      self._path = p.abspath(folders)
      if create:
        if p.isdir(self._path):
          if rm:
            try:
              shutil.rmtree(self._path)
              os.makedirs(self._path)
              cprint(self._path+" recreated")
            except:
              cprint("Unable to remove "+self._path)
        else:
          os.makedirs(self._path)
      else:
        if not p.isdir(self._path):
          raise KeyError('No folder at ' + self._path)
  
  def name(self):
    head, tail = p.split(self._path)
    return tail
    
  def path(self):
    return self._path
    
  def filenames(self):
    f = os.listdir(self._path)
    f.sort()
    return f
    
  def filepaths(self):
    r = []
    for f in self.filenames():
      r.append(p.join(self._path,f))
    return r
    
  def link(self,files):
    for f in files:
      head,tail = p.split(f)
      os.symlink(f,p.join(self._path,tail))
    
  def delete(self):
    shutil.rmtree(self._path,ignore_errors=True)    
    
    
  def copy(self,other):
    for f in other.filepaths():
      head,tail = p.split(f)
      if tail[0] != '.':
        dst = p.join(self.path(),tail)
        if p.islink(f):
          linkto = p.abspath(f+'/.')
          print linkto
          os.symlink(linkto, dst)
        elif p.isdir(f):
          Folder.copy(self[tail],other[tail])
        else:
          shutil.copy(f,dst)
          
  def __getattr__(self,key):
    ''' override dot operator e.g: obj.bla '''
    if key[0] == '_':
      return object.__getattribute__(self,key)

    npy = p.join(self._path,key+'.npy') 
    if p.isfile(npy):
      arr = np.load(npy)
      return arr
      
    pic = p.join(self._path,key+'.p')
    if p.isfile(pic):
      f = open( pic, "rb" )
      #fcntl.flock(f,fcntl.LOCK_SH)
      r = pickle.load( f )
      #fcntl.flock(f,fcntl.LOCK_UN)
      f.close()
      return r
    
    directory = p.join(self._path,key)
    if p.isdir(directory):
      return Folder(directory,self._create) 
    
    if self._create:
      return Folder(p.join(self._path,key),True)
      
    raise KeyError(str(key))
    #return None

  
  def __contains__(self,key):
    ''' override "in" operator e.g.: 'filex' in folder '''
    # TODO: abstract from supported file types
    fil = p.join(self._path,key)  
    npy = p.join(self._path,key+'.npy') 
    pic = p.join(self._path,key+'.p')
    
    return p.isfile(npy) or p.isfile(pic) or p.isdir(fil)
    
    
  def __setattr__(self,key,value):
    ''' override dot operator e.g.: obj.bla = something '''
    # TODO: use non-blocking and "atomic" write
  
    if key[0] == '_':
      object.__setattr__(self,key,value)
    elif type(value) is np.ndarray:
      with FaultTolerantFile(p.join(self._path,key+'.npy')) as f:
        np.save(f,value)
    elif type(value) is Folder:
      # TODO: copy option
      # move all files and remove dir
      fol = Folder(p.join(self._path,key),create=value._create)
      f = os.listdir(value._path)
      for el in f:
        shutil.move(p.join(value._path,el),p.join(fol._path,el))
      os.rmdir(value._path)
    else:
      with FaultTolerantFile(p.join(self._path,key+'.p')) as f:
        pickle.dump(value,f)
      
  
  def __getitem__(self, key):
    ''' override brackets operator e.g.: obj[3] or obj['bla'] '''
    if type(key) is str:
      return Folder.__getattr__(self,key)
    else:
      f = self.filenames()
      files = f[key]
      if type(files) is list:
        res = []
        for f in files:
          name,ext = p.splitext(f)
          res.append(Folder.__getattr__(self,name))
        return res
      else:
        name,ext = p.splitext(files)
        return Folder.__getattr__(self,name)
      
  # override brackets operator e.g.: obj['bla'] = Folder()
  def __setitem__(self, key, value):
    return self.__setattr__(key,value)
    
  # usage: len(obj) -> number of files in folder
  def __len__(self):
    return len(os.listdir(self._path))
    
      
  
if __name__ == "__main__":
  print "test"


import matplotlib
matplotlib.use("Agg")
import warnings
warnings.filterwarnings("ignore", module="matplotlib")

from IPython.display import display
from ipywidgets import widgets 
import signal
import matplotlib.pyplot as plt

import time
import thread
import tensorboard
from tensorboard import Tensorboard
import experiment
import cStringIO
import os

import __init__ as ezex
from ezex import Folder

style_hlink = '<style>.hlink{padding: 5px 10px 5px 10px;display:inline-block;}</style>'

exps = ezex.exfolder

def dashboard(max=8):
  here = Folder('.')
  
  def killtb():
    if 'tb' in here:
      try:
        here.tb.kill()
      except:
        pass

  killtb()
  
  class exp_view:
    def __init__(self,e):
      self.e = e
      #self.name = widgets.Button(description=e.name())
      bname = widgets.HTML(style_hlink+
        '<a class=hlink target="_blank"'+
        'href="http://localhost:'+str(ezex.config['ip']) +'/tree/'+e.name()+'"> '+
        e.name() + ' </a> ')

      self.run_type = widgets.Button()
      self.run_status = widgets.Button()

      space = widgets.Button(description='     ')

      killb = widgets.Button(description='kill')
      delb = widgets.Button(description='delete')
      killb.on_click(lambda _,e=e: experiment.kill(e.path()))

      def delf(_,self=self):
        self.delete()
        experiment.delete(self.e.path())
      delb.on_click(delf)


      tbb = widgets.Button(description='tensorboard')
      tbbb = widgets.HTML(style_hlink+'<a class=hlink target="_blank" href="http://localhost:'+ str(ezex.config['tb']) +'"> (open) </a> ')
      
      def ontb(_,self=self):
        #folder = tensorboard.tbfolder([self.e])
        folder = self.e.path()
        killtb()
        tb = Tensorboard(folder,port=ezex.config['tb'],stdout=True)
        here.tb = tb
        # tb.openbrowser()
      tbb.on_click(ontb)

      self.bar = widgets.HBox((bname,self.run_type,self.run_status,space,tbb,tbbb,space,killb,delb))
      self.plot = widgets.Image(format='png')
      self.view = widgets.VBox((self.bar,self.plot,widgets.HTML('<br><br>')))

      self.th_stop = False
      def loop_fig(self=self):
        while not self.th_stop:
          try:
            # update plot
            try:
              r = self.e.test_r
              i = self.e.test_i
            except:
              r = 0
              i = 0
            #f = plt.figure()
            #f = Figure()
            f,ax = plt.subplots()
            f.set_size_inches((15,2.5))
            f.set_tight_layout(True)
            ax.plot(i,r)
            sio = cStringIO.StringIO()
            f.savefig(sio, format='png',dpi=60)

            self.plot.value = sio.getvalue()

            sio.close()
            plt.close(f)
          except:
            pass #print ex

      self.th = thread.start_new_thread(loop_fig,())

    def update(self):
      try:
        # update labels
        x = experiment.xread(self.e.path())
        self.run_type.description = x['run_type']
        
        # alive?
        try:
          mtime = os.path.getmtime(self.e.path()+'/test_r.npy')
        except OSError:
          mtime = time.time()
        
        if time.time()-mtime > 10*60: # heartbeat 10 min
          self.run_status.description = 'dead'
        else:
          self.run_status.description = x['run_status']
      except:
        pass

    def delete(self):
      self.th_stop = True
  

  main_view = widgets.VBox()
  #display(widgets.HTML('<a target="_blank" href="http://localhost:'+ str(ezex.config['tb']) +'">fjdk</a>'))
  display(main_view)
  
  def loop():
    views = {}
    while True:
      try:
        views2 = {}
        todisplay = []

        i = 0
        for e in reversed(exps[:]):
          if i == max: break
          i = i+1
          v = views[e.name()] if e.name() in views else exp_view(e)
          v.update()
          todisplay = todisplay + [v.view]
          views2[e.name()] = v

        main_view.children = todisplay
        views = views2
        time.sleep(0.5)
      except Exception as ex:
          pass
          #print ex

  th = thread.start_new_thread(loop,())
  


from folder import Folder
import time
from datetime import datetime
import thread
import subprocess
import shutil
import json
import os
import __init__ as ezex
import util

def xwrite(path,data):
  with open(path+'/ezex.json','w') as f:
    json.dump(data,f)

def xread(path):
  with open(path+'/ezex.json') as f:
    return json.load(f)


def create(run_folder,tag=''):
  ''' create unique experiment folder '''
  
  rf = Folder(run_folder)
  
  #basename = 'e'+str(int(time.time()))+'_'+rf.name()
  dstr = datetime.now().strftime('%Y%m%d_%H%M_%S')
  basename =  dstr+'_'+rf.name()+'_'+tag
  name = basename
  i = 1
  while name in ezex.exfolder:
    name = basename + '_' + str(i)
    i = i+1
    if i > 100:
      raise RuntimeError('Could not create unique experiment folder')
      
  # ezex.exp_folder[name] = Folder(create=True)
  # exp = ezex.exp_folder[name]
  # exp.copy(rf)
  path = ezex.config['exfolder']+'/'+name
  os.mkdir(path)
  util.copy(run_folder,path,symlinks=True,ignore='.*')
  return path


def submit(path):
  x = xread(path)
  if ezex.config['scheduler']=='lsf':
    nvd = 'select[nvd]' if x['nvd'] else ''
    cmd = ('bsub -q deflt_centos -M 20000 -W 24:00 -n 1 -R "'+nvd+'" -x '+
      '-oo '+path+'/out_%J '+
      '\''+ 'ezex execute ' + path +'\'')

    print cmd
    out = subprocess.check_output(cmd,shell=True)
    
    import re
    match = re.match('Job <(\d*)>',out)
    if match:
      jid = match.group(1)
    else:
      raise RuntimeError()

  elif ezex.config['scheduler']=='slurm':
    # TODO
    raise RuntimeError('SLURM not yet supported')
  else:
    raise RuntimeError('No scheduling system (e.g. SLURM) present')

  x['job_id'] = jid
  xwrite(path,x)


def execute(path,deltime=7*60):
  x = xread(path)
  t_start = time.time()
  try:
    cmd = x['cmd']
    x['run_status'] = 'running'
    xwrite(path,x)
    print cmd
    os.system(cmd)
    x['run_status'] = 'finished'
  except:
    x['run_status'] = 'aborted'
  finally:
    elapsed = time.time() - t_start
    if elapsed <= deltime:
      delete(path)

    x['end_time'] = time.time()
    xwrite(path,x)


def kill(path):
  try:
    x = xread(path)
    if ezex.config['scheduler']=='lsf':
      id = x['job_id']
      cmd = 'bkill '+str(id)
      out = subprocess.check_output(cmd,shell=True)
    else:
      return False
  except Exception as ex:
    return False


def delete(path):
  # TODO: simplify!
  kill(path)
  def deletet(path=path):
    t = time.time()
    suc = False
    exc = None
    def errorRemoveReadonly(func, path, exc):
      import errno
      import stat
      excvalue = exc[1]
      if func in (os.rmdir, os.remove) and excvalue.errno == errno.EACCES:
        # change the file to be readable,writable,executable: 0777
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  
        # retry
        func(path)
      else:
        exc = "Deletion of "+path+" not possible"
        
    while (time.time()-t < 2):
      try:
        shutil.rmtree(path,ignore_errors=False,onerror=errorRemoveReadonly) 
        suc = not os.path.isdir(path)
        if suc: break
      except Exception as ex:
        exc = ex

      time.sleep(0.2)

    if not suc:
      print exc

  thread.start_new_thread(deletet,())


import subprocess
import os
import time
import signal
import webbrowser
import random
import thread
from folder import Folder

def get_pid(port):
  for lsof in ["lsof","/usr/sbin/lsof"]:
    try:        
      pid = int(subprocess.check_output([lsof,"-t","-i:"+str(port)]))
      break
    except:
      pid = -1
  return pid


def tbfolder(experiments):
  ''' create temporary folder with a subset of experiment folders and tensorboard event files files for tensorboard '''
  
  tbf = Folder(folders=None, create=True) # temporary folder
  for f in experiments:
    tbf[f.name()].link(f.tb.filepaths())
  return tbf.path()


class Tensorboard:
  def __init__(self,logdir,host="127.0.0.1", port=None,force=True,stdout=False):
    ''' start tensorboard process reliably '''
    
    while port is None:
      port = random.randint(6006,6999)
      port = port if get_pid(port) == -1 else None
      
    self.port = port
    self.host = host
    
    #tbpath=os.path.dirname(tensorflow.__file__)+'/tensorboard/tensorboard.py'
    
    # free port
    if force:
      pid = get_pid(port)
      if pid != -1:
        os.kill(pid,signal.SIGTERM)
        print("Tensorboard: Killed process " + str(pid) + " to free port " + str(port))
    
    # launch tensorboard
    '''
    if stdout:
       DEVNULL = sys.stdout
    else:
       DEVNULL = open(os.devnull, 'wb')
   '''
    
    def run(self=self):
      self.output = None
      
      try:
        self.proc = subprocess.Popen(['tensorboard',
                                       "--host=" + self.host + " ",
                                       "--port=" + str(self.port) + " ",
                                       "--logdir=" + logdir],
                                       preexec_fn=os.setsid,
                                       shell=False,
                                       stderr=subprocess.STDOUT)
        self.proc.communicate()            
      except subprocess.CalledProcessError as err:
        self.output = err.output
      except Exception as ex:
        self.output = str(ex)          

    thread.start_new_thread(run,())
                           
    # check if running     
    time.sleep(2)
    if self.output is not None:
      raise RuntimeError("Tensorboard failed on port "+self.port+"\n"+"Output:\n"+self.ouput.getvalue())
    else:
      pass
      #print("Tensorboard started on port "+str(port))
    
    
  def openbrowser(self):
    webbrowser.open_new_tab('http://localhost:'+str(self.port))
    
  def kill(self):
    ''' kill tensorboard process '''
    os.kill(- self.proc.pid,signal.SIGTERM)
    #print("Tensorboard stopped")   
    



from tensorboard import Tensorboard
import __init__ as ex
from __init__ import Folder

if __name__ == '__main__':
  exps = ex.exp_folder[:]
  folder = ex.tbfolder(exps[-1:])
  port=6005
  tb = Tensorboard(folder,port=port)
  tb.openbrowser()


import os
import argparse
import subprocess
import ezex
import experiment
import util

parser = argparse.ArgumentParser(description='ezex is a lightweigt destributed experimentation framework')
subparsers = parser.add_subparsers(dest='command')

p_vis = subparsers.add_parser('dashboard', help='open jupyter notebook dashboard')
p_vis.add_argument('-remote',action='store_true')
p_vis.add_argument('-db',type=int,default=6007)
p_vis.add_argument('-ip',type=int,default=6008)
p_vis.add_argument('-tb',type=int,default=6009)

p_rem = subparsers.add_parser('remote', help='dashboard on a remote maching via ssh port forwards')
p_rem.add_argument('host')
p_rem.add_argument('user')
#p_rem.add_argument('--password',default=None)
p_rem.add_argument('-db',default='6007')
p_rem.add_argument('-ip',default='6008')
p_rem.add_argument('-tb',default='6009')

p_run = subparsers.add_parser('run', help='run experiments (i.e. python scripts)')
p_run.add_argument('path',help='path to python script')
p_run.add_argument('tag',default='')
p_run.add_argument('-l',action='store_true',help='run on this machine')
p_run.add_argument('-nvd',action='store_true',help='run on Nvidia-Node')
p_run.add_argument('-preload',default=None,help='LD_PRELOAD=<PRELOAD> (path relative to script folder)')
p_run.add_argument('-prerun',default=None,help='run this command before the main script')

p_set = subparsers.add_parser('set', help='set global ezex variables (e.g. experiment folder)')
p_set.add_argument('-exfolder',help='path to experiment folder')

p_exe = subparsers.add_parser('execute', help='locally run python script (internal)')
p_exe.add_argument('path',help='path to experiment')

args = parser.parse_args()


if args.command == 'set':
	c = ezex.cread()
	if args.exfolder:
		if not os.path.isdir(args.exfolder):
			os.mkdir(args.exfolder)
		c['exfolder'] = args.exfolder
	ezex.cwrite(c)


if args.command == 'dashboard':
	print 'Experiment root folder at: ' + ezex.config['exfolder']
	print 'containing ' + str(len(ezex.exfolder[:])) + ' experiments'
	print ''

	util.free_port(args.db)
	util.free_port(args.ip)
	util.free_port(args.tb)

	c = ezex.cread()
	c['db'] = args.db
	c['ip'] = args.ip
	c['tb'] = args.tb
	ezex.cwrite(c)

	proc = subprocess.Popen(('jupyter notebook --port='+str(args.ip) +' '+c['exfolder']).split(' '))

	f = ezex.home+'/dashboard.ipynb'
	os.system('jupyter notebook --port='+str(args.db) +' '+f)
	
if args.command == 'remote':
	import os
	import argparse
	import ezex
	import thread
	import webbrowser

	webbrowser.open_new_tab('http://localhost:'+str(args.db)+'/tree/dashboard.ipynb')

	util.free_port(args.db)
	util.free_port(args.ip)
	util.free_port(args.tb)
	
	cmd = ('ezex dashboard -remote' +
					 ' -db ' + str(args.db) +
					 ' -ip ' + str(args.ip) +
					 ' -tb ' + str(args.tb) )

	os.system("ssh -c arcfour "+args.user+"@"+args.host+
						" -L "+args.db+":localhost:"+args.db+
						" -L "+args.ip+":localhost:"+args.ip+
						" -L "+args.tb+":localhost:"+args.tb+
						" '"+cmd+"'")


if args.command == 'run':
	run_folder = os.path.dirname(args.path)
	run_script = os.path.basename(args.path)
	path = experiment.create(run_folder,args.tag)

	preload = '' if not args.preload else 'LD_PRELOAD='+ args.preload
	prerun = '' if not args.prerun else args.prerun+';'
	cmd = '/bin/bash -c "'+ prerun +'cd '+ path +';'+ preload +' python '+run_script+'"'

	experiment.xwrite(path,{'run_type':'local' if args.l else 'job',
													'run_status':'pending',
													'nvd':args.nvd,
													'cmd':cmd})
	if args.l:
	  print 'run local'
	  experiment.execute(path)
	else:
	  print 'submit job'
	  experiment.submit(path)


if args.command == 'execute':
	experiment.execute(args.path)



from folder import Folder
import os
import json
import shutil
home = os.path.dirname(__file__)

def cwrite(data):
  with open(home+'/config.json','w') as f:
    json.dump(data,f)

def cread():
  with open(home+'/config.json') as f:
    return json.load(f)

if not os.path.isfile(home+'/config.json'):
	shutil.copy(home+'/default/config.json', home+'config.json')

if not os.path.isfile(home+'/dashboard.ipynb'):
	shutil.copy(home+'/default/dashboard.ipynb', home+'dashboard.ipynb')

config = cread()
if not config.has_key('exfolder'):
  config['exfolder'] = home+'/experiments'
if not config.has_key('scheduler'):
  config['scheduler'] = 'lsf'
cwrite(config)

exfolder = Folder(config['exfolder'],create=False)




from setuptools import setup

setup(name='ezex',
      version='0.1',
      description='lightweigt destributed experimentation framework',
      url='https://github.com/SimonRamstedt/ezex',
      author='Simon Ramstedt',
      author_email='simonramstedt@gmail.com',
      license='MIT',
      packages=['ezex'],
      scripts=['bin/ezex'],
      install_requires=['jupyter'],
      zip_safe=False)


import os
import shutil
import stat


def copy(src, dst, symlinks = False, ignore = None):
  ign = shutil.ignore_patterns(ignore)
  copytree(src,dst,symlinks,ign)

def copytree(src, dst, symlinks = False, ignore = None):
  if not os.path.exists(dst):
    os.makedirs(dst)
    shutil.copystat(src, dst)
  lst = os.listdir(src)
  if ignore:
    excl = ignore(src, lst)
    lst = [x for x in lst if x not in excl]
  for item in lst:
    s = os.path.join(src, item)
    d = os.path.join(dst, item)
    if symlinks and os.path.islink(s):
      if os.path.lexists(d):
        os.remove(d)
      os.symlink(os.readlink(s), d)
      try:
        st = os.lstat(s)
        mode = stat.S_IMODE(st.st_mode)
        os.lchmod(d, mode)
      except:
        pass # lchmod not available
    elif os.path.isdir(s):
      copytree(s, d, symlinks, ignore)
    else:
      shutil.copy2(s, d)

def free_port(port):
  import signal
  import subprocess

  for lsof in ["lsof","/usr/sbin/lsof"]:
    try:        
      out = subprocess.check_output([lsof,"-t","-i:"+str(port)])
      for l in out.splitlines():
        pid = int(l)
        os.kill(pid,signal.SIGTERM)
        print("Killed process " + str(pid) + " to free port " + str(port))

      break
    except subprocess.CalledProcessError:
      pid = -1
    except OSError:
      pass